/*
** Module   : datetime.java
** Abstract : Progress 4GL compatible datetime object
**
** Copyright (c) 2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --------------------Description--------------------------
** 001 OM  20130111 Initial commit with basic functionality.
*/
package com.goldencode.p2j.util;

import java.util.*;

/**
 * A class that represents a Progress 4GL compatible datetime object whose
 * data is mutable.  All Progress language features related to date data
 * types are supported including all function that can be accessed via the
 * date aware operators and built-in Progress date-related functions.
 * <p>
 * This class extends the <code>date</code> class so all methods should work
 * directly. If they don't they will be overridden. Also should the following
 * static implementations:
 * <p>
 * <pre>
 * +        operator (datetime + integer millis) {@link #plusDays(date, integer)} 
 * +        operator (datetime + decimal millis) {@link #plusDays(date, decimal)}
 * -        operator (datetime - integer millis) {@link #minusDays(date, integer)}
 * -        operator (datetime - decimal days)   {@link #minusDays(date, decimal)}
 * -        operator (datetime - date)           {@link #differenceNum}
 * =  or EQ operator                       {@link CompareOps#isEqual}
 * &lt;&gt; or NE operator                 {@link CompareOps#isNotEqual}
 * &gt;  or GT operator                    {@link CompareOps#isGreaterThan}
 * &gt;= or GE operator                    {@link CompareOps#isGreaterThanOrEqual}
 * &lt;  or LT operator                    {@link CompareOps#isLessThan}
 * &lt;= or LE operator                    {@link CompareOps#isLessThanOrEqual}
 * day      function                       {@link #day}
 * weekday  function                       {@link #weekday}
 * month    function                       {@link #month}
 * year     function                       {@link #year}
 * maximum  function                       {@link CompareOps#maximum(BaseDataType[])}
 * minimum  function                       {@link CompareOps#minimum(BaseDataType[])}
 * decimal  function (to convert dates)    {@link #decimalValue}
 * integer  function (to convert dates)    {@link #integerValue}
 * string   function (default format)      {@link #toString()}
 *                   (user-defined format) {@link #toString(String)} 
 *                   (export format)       {@link #toStringExport()}
 * time     function                       {@link #secondsSinceMidnight}
 * etime    function                       {@link #elapsed}
 * now      function                       {@link #datetime()}
 * datetime function (integer parameter)   {@link #datetime(NumberType)}
 * datetime function (integer m, d, y)     {@link #datetime(NumberType, NumberType, NumberType)}
 * datetime function (string parameter)    {@link #datetime(String)}
 * </pre>
 * <p>
 * Default data type display formats: DATETIME 99/99/9999 HH:MM:SS.SSS
 * <p>
 *
 * @author    OM
 */
public class datetime
extends date
{
   /**
    * The time passed since midnight of <code>date.julian<code> in 
    * milliseconds. 
    */
   private int time = 0;

   /** The default format string. */
   private static final String defaultFormat = "99/99/9999";

   /** Mockup implementation for SESSION:TIME-SOURCE attribute. */
   private static String timeSource = "local";
   
   /**
    * Creates an instance that represents current date and time using the 
    * default timezone and the local system clock time.
    */
   public datetime()
   {
      // base class will initialize the
      Calendar now = Calendar.getInstance();
      this.time = now.get(Calendar.MILLISECOND) +
                  1000 * (now.get(Calendar.SECOND) +
                          60 * (now.get(Calendar.MINUTE) +
                                60 * now.get(Calendar.HOUR_OF_DAY)));
   }

   /**
    * Constructs an instance with the exact same state as the passed
    * instance.
    *
    * @param    d
    *           The instance from which to copy state.
    */
   public datetime(date d)
   {
      super(d); // assign date part

      if (d instanceof datetime)
      {
         // then the time, if the case
         time = ((datetime)d).getTime();
      }
   }

   /**
    * Construct a datetime object similar to the 3rd form of datetime 
    * function from 4GL.
    * 
    * @param   month
    *          The month for the new datetime object.
    * @param   day
    *          The day for the new datetime object.
    * @param   year
    *          The year for the new datetime object.
    * @param   hours
    *          The hours for the new datetime object.
    * @param   minutes
    *          The minutes for the new datetime object.
    * @param   seconds
    *          The seconds for the new datetime object.
    * @param   millis
    *          The milliseconds for the new datetime object.
    */
   public datetime(double month,
                   double day,
                   double year,
                   double hours,
                   double minutes,
                   double seconds,
                   double millis)
   {
      // the super class will take care of the date part
      super(month, day, year);

      // setting time
      setTime((int)hours, (int)minutes, (int)seconds, (int)millis);
   }

   /**
    * Construct a datetime object similar to the 3rd form of datetime 
    * function from 4GL.
    * 
    * @param   month
    *          The month for the new datetime object.
    * @param   day
    *          The day for the new datetime object.
    * @param   year
    *          The year for the new datetime object.
    * @param   hours
    *          The hours for the new datetime object.
    * @param   minutes
    *          The minutes for the new datetime object.
    * @param   seconds
    *          The seconds for the new datetime object.
    */
   public datetime(double month,
                   double day,
                   double year,
                   double hours,
                   double minutes,
                   double seconds)
   {
      // the super class will take care of the date part
      super(month, day, year);

      // setting time
      setTime((int)hours, (int)minutes, (int)seconds, 0);
   }

   /**
    * Construct a datetime object similar to the 3rd form of datetime 
    * function from 4GL.
    * 
    * @param   month
    *          The month for the new datetime object.
    * @param   day
    *          The day for the new datetime object.
    * @param   year
    *          The year for the new datetime object.
    * @param   hours
    *          The hours for the new datetime object.
    * @param   minutes
    *          The minutes for the new datetime object.
    */
   public datetime(double month,
                   double day,
                   double year,
                   double hours,
                   double minutes)
   {
      // the super class will take care of the date part
      super(month, day, year);

      // setting time
      setTime((int)hours, (int)minutes, 0, 0);
   }

   /**
    * Construct a datetime object similar to the 3rd form of datetime 
    * function from 4GL.
    * 
    * @param   month
    *          The month for the new datetime object.
    * @param   day
    *          The day for the new datetime object.
    * @param   year
    *          The year for the new datetime object.
    * @param   hours
    *          The hours for the new datetime object.
    * @param   minutes
    *          The minutes for the new datetime object.
    * @param   seconds
    *          The seconds for the new datetime object.
    * @param   millis
    *          The milliseconds for the new datetime object.
    */
   public datetime(NumberType month,
                   NumberType day,
                   NumberType year,
                   NumberType hours,
                   NumberType minutes,
                   NumberType seconds,
                   NumberType millis)
   {
      // the super class will take care of the date part
      super(month, day, year);

      // setting time
      setTime(hours.intValue(), 
              minutes.intValue(), 
              seconds.intValue(), 
              millis.intValue());
   }

   /**
    * Construct a datetime object similar to the 3rd form of datetime 
    * function from 4GL.
    * 
    * @param   month
    *          The month for the new datetime object.
    * @param   day
    *          The day for the new datetime object.
    * @param   year
    *          The year for the new datetime object.
    * @param   hours
    *          The hours for the new datetime object.
    * @param   minutes
    *          The minutes for the new datetime object.
    * @param   seconds
    *          The seconds for the new datetime object.
    */
   public datetime(NumberType month,
                   NumberType day,
                   NumberType year,
                   NumberType hours,
                   NumberType minutes,
                   NumberType seconds)
   {
      // the super class will take care of the date part
      super(month, day, year);

      // setting time
      setTime(hours.intValue(), 
              minutes.intValue(), 
              seconds.intValue(), 
              0);
   }

   /**
    * Construct a datetime object similar to the 3rd form of datetime 
    * function from 4GL.
    * 
    * @param   month
    *          The month for the new datetime object.
    * @param   day
    *          The day for the new datetime object.
    * @param   year
    *          The year for the new datetime object.
    * @param   hours
    *          The hours for the new datetime object.
    * @param   minutes
    *          The minutes for the new datetime object.
    */
   public datetime(NumberType month,
                   NumberType day,
                   NumberType year,
                   NumberType hours,
                   NumberType minutes)
   {
      // the super class will take care of the date part
      super(month, day, year);

      // setting time
      setTime(hours.intValue(), minutes.intValue(), 0, 0);
   }

   /**
    * Parses a string with a format based on the Progress 4GL rules for
    * converting a string to a date.
    */
   public datetime(String d)
   throws ErrorConditionException
   {
      /* algorithm: check for time tokens (they are separated by ':')
         cut time token and call super constructor
         parse the time and set accordingly */

      super.instatiateFromStringWorker(d);
   }

   /**
    * Parses a string with a format based on the Progress 4GL rules for
    * converting a string to a date.
    */
   public datetime(character d)
   throws ErrorConditionException
   {
      if (d.isUnknown())
      {
         setUnknown();
         return;
      }
      /* algorithm: check for time tokens (they are separated by ':')
         cut time token and call super constructor
         parse the time and set accordingly */

      super.instatiateFromStringWorker(d.getValue());
  }

   /**
    * Constructs an instance with the exact same state as the passed
    * instance and having the specified time.
    * 
    * @param   d
    *          The date.
    * @param   time
    *          The time from midnight.
    */
   public datetime(date d, integer time)
   {
      super(d);
      if (time.isUnknown())
      {
         setUnknown();
      }
      else
      {
         this.time = time.intValue();
      }
   }

   /**
    * Constructs an instance with the exact same state as the passed
    * instance and having the specified time. Assumes that parameters are 
    * normalized.
    * 
    * @param   days
    *          The number of days (in Progress way).
    * @param   time
    *          The time from midnight.
    */
   public datetime(int days, int time)
   {
      super(days);
      this.time = time;
   }

   /**
    * Returns the number of milliseconds difference between the datetimes of 
    * two instances.
    *
    * @param    op1
    *           The left operand.
    * @param    op2
    *           The right operand (instance to subtract from the left
    *           operand).
    *
    * @return   The difference in days between the two datetimes or 
    *           <code>unknown value</code> if any input is
    *           <code>unknown value</code>.
    */
   public static int64 differenceNum(datetime op1, datetime op2)
   {
      if (op1.isUnknown() || op2.isUnknown())
      {
         return new int64();
      }
      
      return new int64(op1.getAbsoluteTimeOffset() -
                       op2.getAbsoluteTimeOffset());
   }

   /**
    * Obtain the millis from midnight. This is the implementation of 
    * <code>time()</code> function of 4GL.
    * 
    * @return  milliseconds passed from midnight.
    */
   public integer getTimeNum()
   {
      return isUnknown() ? new integer() : new integer(time);
   }

   /**
    * Obtain the millis from midnight.
    * 
    * @return  milliseconds passed from midnight.
    */
   public int getTime()
   {
      return time;
   }

   /**
    * Sets the millis from midnight.
    * 
    * @param   millis
    *          Milliseconds passed from midnight.
    */
   public void setTime(NumberType millis)
   {
      if (millis.isUnknown())
      {
         setUnknown();
      }
      else 
      {
         time = millis.intValue();
      }
   }

   /**
    * Sets the millis from midnight.
    * 
    * @param   millis
    *          Milliseconds passed from midnight.
    */
   public void setTime(int millis)
   {
      time = millis;
   }

   /**
    * Extract the hours time-element from this datetime.
    * 
    * @return the hour of this datetime object
    */
   public int getHours()
   {
      return time / (1000 * 60 * 60);
   }


   /**
    * Extract the hours time-element from this datetime.
    * 
    * @return  the hour of this datetime object or unknown value if this
    *          is an unknown object.
    */
   public integer getHoursNum()
   {
      return isUnknown() ? new integer() : new integer(getHours());
   }


   /**
    * Extract the minutes time-element from this datetime.
    * 
    * @return the minute of this datetime object
    */
   public int getMinutes()
   {
      int timeInMinutes = time / (1000 * 60);
      return timeInMinutes % 60;
   }


   /**
    * Extract the minutes time-element from this datetime.
    * 
    * @return  the minute of this datetime object or unknown value if this
    *          is an unknown object.
    */
   public integer getMinutesNum()
   {
      return isUnknown() ? new integer() : new integer(getMinutes());
   }


   /**
    * Extract the seconds time-element from this datetime.
    * 
    * @return the second of this datetime object
    */
   public int getSeconds()
   {
      int timeInSeconds = time / 1000;
      return timeInSeconds % 60;
   }


   /**
    * Extract the seconds time-element from this datetime.
    * 
    * @return  the second of this datetime object or unknown value if this
    *          is an unknown object.
    */
   public integer getSecondsNum()
   {
      return isUnknown() ? new integer() : new integer(getSeconds());
   }


   /**
    * Extract the milliseconds time-element from this datetime.
    * 
    * @return the millisecond of this datetime object
    */
   public int getMilliseconds()
   {
      return time % 1000;
   }


   /**
    * Extract the milliseconds time-element from this datetime.
    * 
    * @return  the millisecond of this datetime object or unknown value if 
    *          this is an unknown object.
    */
   public integer getMillisecondsNum()
   {
      return isUnknown() ? new integer() : new integer(getMillisecondsNum());
   }


   /**
    * Sets the internal time. This method will validate the parameters and
    * throw <code>IllegalArgumentException</code> exception if they are 
    * incorrect.
    * 
    * @param   hours
    *          Hours to be set.
    * @param   mins
    *          Minutes to be set.
    * @param   secs
    *          Seconds to be set.
    * @param   millis
    *          Milli-seconds to be set.
    */
   public void setTime(int hours, int mins, int secs, int millis)
   {
      if (hours > 23 || hours < 0)
      {
         throw new IllegalArgumentException(
               "Hours argument must be between 0 and 23");
      }

      if (mins > 59 || mins < 0)
      {
         throw new IllegalArgumentException(
               "Minutes argument must be between 0 and 59");
      }

      if (secs > 59 || secs < 0)
      {
         throw new IllegalArgumentException(
               "Seconds argument must be between 0 and 59");
      }

      if (millis > 999 || millis < 0)
      {
         throw new IllegalArgumentException(
               "Hours argument must be between 0 and 999");
      }

      time = millis + 1000 * (secs + 60 * (mins + 60 * hours));
   }

   /**
    * Adds <code>milliseconds</code> to the datetime of the left operand and 
    * returns a new instance representing that datetime.
    *
    * @param   dt
    *          The left operand.
    * @param   millis
    *          The right operand (the number of milliseconds to add. 
    *          Note that this can be a negative number to effectively 
    *          subtract instead of add.
    *
    * @return  The date which is <code>millis</code> different from the
    *          left operand or <code>unknown value</code> if <code>dt</code> 
    *          is <code>unknown value</code>.
    */
   public static datetime plusMillis(datetime dt, long millis)
   {
      if (dt.isUnknown())
      {
         return datetime.instantiateUnknownDatetime();
      }
      long newMillis = dt.time + millis;
      long days = dt.longValue();
      if (newMillis < 0)
      {
         // must go back in time!
         // TODO: OM: here there is a bug, must be fixed
         days -= - newMillis / date.MILLIS_PER_DAY;
         newMillis = - newMillis % date.MILLIS_PER_DAY;
      }
      else if (newMillis >= date.MILLIS_PER_DAY)
      {
         // go forward in time
         days += newMillis / date.MILLIS_PER_DAY;
         newMillis = newMillis % date.MILLIS_PER_DAY;
      }
      return new datetime((int)days, (int)newMillis);
   }

   /**
    * Adds <code>milliseconds</code> to the datetime of the left operand and 
    * returns a new instance representing that datetime.
    *
    * @param   dt
    *          The left operand.
    * @param   millis
    *          The right operand (the number of milliseconds to add. 
    *          Note that this can be a negative number to effectively 
    *          subtract instead of add.
    *
    * @return  The date which is <code>millis</code> different from the
    *          left operand or <code>unknown value</code> if any input is
    *          <code>unknown value</code>.
    */
   public static datetime plusMillis(datetime dt, int64 millis)
   {
      return (dt.isUnknown() || millis.isUnknown()) ?
              datetime.instantiateUnknownDatetime() :
              plusMillis(dt, millis.longValue());
   }

   /**
    * Adds <code>milliseconds</code> to the datetime of the left operand and 
    * returns a new instance representing that datetime.
    *
    * @param   dt
    *          The left operand.
    * @param   millis
    *          The right operand (the number of milliseconds to add. 
    *          Note that this can be a negative number to effectively 
    *          subtract instead of add.
    *
    * @return  The date which is <code>millis</code> different from the
    *          left operand or <code>unknown value</code> if any input is
    *          <code>unknown value</code>.
    */
   public static datetime plusMillis(datetime dt, integer millis)
   {
      return (dt.isUnknown() || millis.isUnknown()) ?
              datetime.instantiateUnknownDatetime() :
              plusMillis(dt, millis.longValue());
   }

   /**
    * Subtracts <code>milliseconds</code> to the datetime of the left operand
    * and returns a new instance representing that datetime.
    *
    * @param   dt
    *          The left operand.
    * @param   millis
    *          The right operand (the number of milliseconds to subtract. 
    *          Note that this can be a negative number to effectively 
    *          add instead of subtract.
    *
    * @return  The date which is <code>millis</code> different from the
    *          left operand or <code>unknown value</code> if <code>dt</code> 
    *          is <code>unknown value</code>.
    */
   public static datetime minusMillis(datetime dt, long millis)
   {
      return (dt.isUnknown()) ?
              datetime.instantiateUnknownDatetime() :
              plusMillis(dt, -millis);
   }

   /**
    * Subtracts <code>milliseconds</code> to the datetime of the left operand
    * and returns a new instance representing that datetime.
    *
    * @param   dt
    *          The left operand.
    * @param   millis
    *          The right operand (the number of milliseconds to subtract. 
    *          Note that this can be a negative number to effectively 
    *          add instead of subtract.
    *
    * @return  The date which is <code>millis</code> different from the
    *          left operand or <code>unknown value</code> if any input is
    *          <code>unknown value</code>.
    */
   public static datetime minusMillis(datetime dt, int64 millis)
   {
      return (dt.isUnknown() || millis.isUnknown()) ?
              datetime.instantiateUnknownDatetime() :
              plusMillis(dt, -millis.longValue());
   }

   /**
    * Subtracts <code>milliseconds</code> to the datetime of the left operand
    * and returns a new instance representing that datetime.
    *
    * @param   dt
    *          The left operand.
    * @param   millis
    *          The right operand (the number of milliseconds to subtract. 
    *          Note that this can be a negative number to effectively 
    *          add instead of subtract.
    *
    * @return  The date which is <code>millis</code> different from the
    *          left operand or <code>unknown value</code> if any input is
    *          <code>unknown value</code>.
    */
   public static datetime minusMillis(datetime dt, integer millis)
   {
      return (dt.isUnknown() || millis.isUnknown()) ?
              datetime.instantiateUnknownDatetime() :
              plusMillis(dt, -millis.longValue());
   }


   /**
    * Returns the time in milliseconds (MTIME, similar to TIME, which 
    * returns seconds since midnight).
    * 
    * @return  the time in milliseconds.
    */
   public static integer millisecondsSinceMidnight()
   {
      return new integer(new datetime().time);
   }

   /**
    * Returns the time in milliseconds (MTIME, similar to TIME, which 
    * returns seconds since midnight).
    * 
    * @param   dt
    *          the datetime or datetime-tz to evaluate.
    *
    * @return  the time in milliseconds.
    */
   public static integer millisecondsSinceMidnight(datetime dt)
   {
      return dt.isUnknown() ? new integer() : new integer(dt.time);
   }

   /**
    * Creates a new <code>datetime</code> instance that represents the 
    * <code>unknown value</code>.
    * 
    * @return   An instance that represents the <code>unknown value</code>.
    */
   public static datetime instantiateUnknownDatetime()
   {
      datetime ret = new datetime();
      ret.setUnknown();
      return ret;
   }

   /**
    * Creates a new instance of the same type that represents the 
    * <code>unknown value</code>.
    *
    * @return   An instance that represents the <code>unknown value</code>.
    */
   public BaseDataType instantiateUnknown()
   {
      return instantiateUnknownDatetime();
   }

   /**
    * Creates a new instance of the same type that represents the 
    * default initialized value.
    *
    * @return   An instance that represents the default value.
    */
   public BaseDataType instantiateDefault()
   {
      return instantiateUnknownDatetime();
   }

   /**
    * Return the default display format string for this type.
    *
    * @return   The default format string.
    */
   public String defaultFormatString()
   {
      return defaultFormat;
   }   

   /**
    * Compares this instance with the specified instance and returns a -1
    * if this instance is less than the specified, 0 if the two instances
    * are equal and 1 if this instance is greater than the specified
    * instance.  This is the implementation of the <code>Comparable</code>
    * interface.
    * <p>
    * The algorithm will fail to give meaningful results in the case where
    * one tries to sort against other objects that do not represent compatible
    * values.
    *
    * @param    obj
    *           The date instance to compare against.
    *
    * @return   -1, 0 or 1 depending on if this instance is less than,
    *           equal to or greater (respectively) than the specified
    *           instance <code>obj</code>.
    */
   public int compareTo(Object obj)
   {
      date that = null;
      
      if (obj instanceof date)
      {
         that = (date) obj;
      }
      else
      {
         return -1;
      }

      if (this.isUnknown() || that.isUnknown())
      {
         //TODO: decide if two unknowns are equals
         return -1;
      }

      int result = 0;
      long thisAbs = this.getAbsoluteTimeOffset();
      long thatAbs = that.getAbsoluteTimeOffset();
      
      if (thisAbs < thatAbs)
      {
         result = -1;
      }
      else if (thisAbs > thatAbs)
      {
         result = 1;
      }

      return result;
   }

   /**
    * An equality test which handles the most common cases with the least
    * amount of overhead.  The superclass implementation is invoked otherwise.
    * <p>
    * This implementation is consistent with {@link #hashCode()}.
    * 
    * @see com.goldencode.p2j.util.BaseDataType#equals(java.lang.Object)
    */
   public boolean equals(Object o)
   {
      if (this == o)
      {
         return true;
      }
      
      if (o instanceof datetime)
      {
         datetime that = (datetime) o;
         
         boolean u1 = this.isUnknown();
         boolean u2 = that.isUnknown();
         
         if (u1 && u2)
         {
            return true;
         }
         
         if (u1 || u2)
         {
            return false;
         }

         return (this.longValue() == that.longValue()) && 
                (this.time == that.time);
      }
      
      return super.equals(o);
   }

   /**
    * Hash code implementation which is consistent with {@link
    * BaseDataType#equals}.
    *
    * @return  Hash code value for this object instance.
    */
   public int hashCode()
   {
      int result = 17;
      if (isUnknown())
      {
         return result;
      }
      
      result = 37 * result + (int) (longValue() ^ (longValue() >>> 32));
      result = 37 * result + time;
      
      return result;
   }

   /**
    * Does the same as standard <code>clone()</code> method but returns an
    * instance of <code>BaseDataType</code> and doesn't throw the
    * <code>CloneNotSupportedException</code>.
    *
    * @return   A clone of this instance.  
    */
   public BaseDataType duplicate()
   {
      return new datetime(this);
   }

   /**
    * Returns a character representation of a <code>DATE</code>, 
    * <code>DATETIME</code> or <code>DATETIME-TZ</code> that conforms to the 
    * ISO 8601 standard for date/time representations.
    * 
    * @return  A string that is the ISO representation of this date
    *          (YYYY-MM-DDTHH:MM:SS.SSS)
    */
   protected String getIsoDate()
   {
      if (isUnknown())
      {
         return "?";
      }
      
      StringBuilder sb = new StringBuilder(super.getIsoDate());
      sb.append("T");
      sb.append(date.formatNDigits(getHours(), 2));
      sb.append(":");
      sb.append(date.formatNDigits(getMinutes(), 2));
      sb.append(":");
      sb.append(date.formatNDigits(getSeconds(), 2));
      sb.append(".");
      sb.append(date.formatNDigits(getMilliseconds(), 3));

      return sb.toString();
   }

   public static character getTimeSource()
   {
      return new character(datetime.timeSource);
   }

   /**
    * Specifies the client or database server machine that serves as the 
    * time source for applications running during the P2J session.
    * <code>TIME-SOURCE</code> accepts either of the following settings:
    * <ul>
    *    <li>"local" or the null string ("") — Your application uses the 
    *       client machine as its time source. The default value is "local".
    *    <li>"dbname" — Your application uses the machine running the server 
    *    for the database with the name dbname as its time source.
    * </ul>
    * 
    * @param   timeSource
    *          The time source to be set
    */
   public static void setTimeSource(String timeSource)
   {
      System.err.println("TIME-SOURCE = " + timeSource);
      datetime.timeSource = timeSource;
   }
   
   /**
    * Specifies the client or database server machine that serves as the 
    * time source for applications running during the P2J session.
    * <code>TIME-SOURCE</code> accepts either of the following settings:
    * <ul>
    *    <li>"local" or the null string ("") — Your application uses the 
    *       client machine as its time source. The default value is "local".
    *    <li>"dbname" — Your application uses the machine running the server 
    *    for the database with the name dbname as its time source.
    * </ul>
    * 
    * @param   timeSource
    *          The time source to be set
    */
   public static void setTimeSource(character timeSource)
   {
      if (timeSource != null && timeSource.isUnknown())
      {
         // ErrorManager.recordOrThrowError(NNN, message);
      }
      setTimeSource(timeSource == null ? "" : timeSource.getValue());
   }

   /**
    * Returns the absolute time in milliseconds since Postgres 'Epoch' or 0
    * if this is unknown.
    * @return  absolute time in milliseconds
    */
   @Override
   protected long getAbsoluteTimeOffset()
   {
      return isUnknown() ? 0 : (super.getAbsoluteTimeOffset() + time);
   }
}

