/*
** Module   : datetimetz.java
** Abstract : Progress 4GL compatible datetime-tz object
**
** Copyright (c) 2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --------------------Description--------------------------
** 001 OM  20130111 Initial commit with basic functionality.
*/
package com.goldencode.p2j.util;

/**
 * A class that represents a Progress 4GL compatible datetime-tz object whose
 * data is mutable.  All Progress language features related to date data
 * types are supported including all function that can be accessed via the
 * date aware operators and built-in Progress date-related functions.
 * <p>
 * This class extends the <code>date</code> class so all methods should work
 * directly. If they don't they will be overridden. Also should the following
 * static implementations:
 * <p>
 * <pre>
 * +        operator (date + integer days) {@link #plusDays(date,integer)} 
 * +        operator (date + decimal days) {@link #plusDays(date,decimal)}
 * -        operator (date - integer days) {@link #minusDays(date,integer)}
 * -        operator (date - decimal days) {@link #minusDays(date,decimal)}
 * -        operator (date - date)         {@link #differenceNum}
 * =  or EQ operator                       {@link CompareOps#isEqual}
 * &lt;&gt; or NE operator                       {@link CompareOps#isNotEqual}
 * &gt;  or GT operator                       {@link CompareOps#isGreaterThan}
 * &gt;= or GE operator                       {@link CompareOps#isGreaterThanOrEqual}
 * &lt;  or LT operator                       {@link CompareOps#isLessThan}
 * &lt;= or LE operator                       {@link CompareOps#isLessThanOrEqual}
 * day      function                       {@link #day}
 * weekday  function                       {@link #weekday}
 * month    function                       {@link #month}
 * year     function                       {@link #year}
 * maximum  function                       {@link CompareOps#maximum(BaseDataType[])}
 * minimum  function                       {@link CompareOps#minimum(BaseDataType[])}
 * decimal  function (to convert dates)    {@link #decimalValue}
 * integer  function (to convert dates)    {@link #integerValue}
 * string   function (default format)      {@link #toString()}
 *                   (user-defined format) {@link #toString(String)} 
 *                   (export format)       {@link #toStringExport()}
 * time     function                       {@link #secondsSinceMidnight}
 * etime    function                       {@link #elapsed}
 * today    function                       {@link #date()}
 * date     function (integer parameter)   {@link #date(NumberType)}
 * date     function (integer m, d, y)     {@link #date(NumberType,NumberType,NumberType)}
 * date     function (string parameter)    {@link #date(String)}
 * </pre>
 * <p>
 * Default data type display formats: DATETIME-TZ 99/99/9999 HH:MM:SS.SSS+HH:MM
 * <p>
 * TODO: OM: This class also features a built-in regression testing harness that is
 * invoked via {@link #main}.
 *
 * @author    OM
 */
public class datetimetz
extends datetime
{
   private int tz = 0;

   public datetimetz()
   {
    //TODO: OM: implement this correctly
   }

   public datetimetz(String val)
   {
      super(val);
      //TODO: OM: implement this correctly
   }

   /**
    * Creates a new <code>datetimetz</code> instance that represents the 
    * <code>unknown value</code>.
    * 
    * @return   An instance that represents the <code>unknown value</code>.
    */
   public static datetimetz instantiateUnknownDatetimetz()
   {
      datetimetz ret = new datetimetz();
      ret.setUnknown();
      return ret;
   }

   /**
    * Creates a new instance of the same type that represents the 
    * <code>unknown value</code>.
    *
    * @return   An instance that represents the <code>unknown value</code>.
    */
   public BaseDataType instantiateUnknown()
   {
      return instantiateUnknownDatetimetz();
   }

   /**
    * Creates a new instance of the same type that represents the 
    * default initialized value.
    *
    * @return   An instance that represents the default value.
    */
   public BaseDataType instantiateDefault()
   {
      return instantiateUnknownDatetimetz();
   }

   /**
    * Returns a character representation of a <code>DATE</code>, 
    * <code>DATETIME</code> or <code>DATETIME-TZ</code> that conforms to the 
    * ISO 8601 standard for date/time representations.
    * 
    * @return  A string that is the ISO representation of this date
    *          (YYYY-MM-DDTHH:MM:SS.SSS+HH:MM)
    */
   protected String getIsoDate()
   {
      if (isUnknown())
      {
         return "?";
      }

      StringBuilder sb = new StringBuilder(super.getIsoDate());
      sb.append("+");
      sb.append(date.formatNDigits(getTzHours(), 2));
      sb.append(":");
      sb.append(date.formatNDigits(getTzMinutes(), 2));

      return sb.toString();
   }

   /**
    * Obtains the minutes of timezone offset.
    * 
    * @return  The minutes of timezone offset.
    */
   public int getTzMinutes()
   {
      return tz % 60;
   }

   /**
    * Obtains the hours of timezone offset.
    * 
    * @return  The hours of timezone offset or unknown value if this 
    *          datetime-tz is unknown.
    */
   public int getTzHours()
   {
      return tz / 60;
   }

   /**
    * Obtains the minutes of timezone offset.
    * 
    * @return  The minutes of timezone offset or unknown value if this 
    *          datetime-tz is unknown.
    */
   public integer getTzMinutesNum()
   {
      return isUnknown() ? new integer() : new integer(getTzMinutes());
   }

   /**
    * Obtains the hours of timezone offset.
    * 
    * @return  The hours of timezone offset or unknown value if this 
    *          datetime-tz is unknown.
    */
   public integer getTzHoursNum()
   {
      return isUnknown() ? new integer() : new integer(getTzHours());
   }
}
